#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-05-21
# File  : unbound_status.py

# 2024-09-10: fixed crash if unbound is not running

from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from re import match as re_match
from typing import Any

from cmk.utils import debug
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
    StringTable,
)

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Result,
    Service,
    State,
    check_levels,
    register,
    render,
)


@dataclass(frozen=True)
class UnboundStatus:
    status: str
    modules: Sequence[str] | None = None
    options: Sequence[str] | None = None
    pid: int | None = None
    threads: int | None = None
    uptime: int | None = None
    verbosity: int | None = None
    version: str | None = None

    @classmethod
    def parse(cls, string_table: StringTable):
        # set defaults
        modules = None
        options = None
        pid = None
        status = None
        threads = None
        uptime = None
        verbosity = None
        version = None

        for line in string_table:
            key, value = line[0].split(' ', 1)
            key = key.rstrip(':')
            match key:
                case 'version':
                    version = str(value)
                case 'verbosity':
                    verbosity = int(value)
                case 'threads':
                    threads = int(value)
                case 'modules':
                    # 3 [ subnet validator iterator ]
                    modules = value.split(' [ ')[-1].strip(']').strip(' ').split(' ')
                case 'uptime':
                    # 577 seconds
                    uptime = int(value.split(' ')[0])
                case 'options':
                    # reuseport control(ssl)
                    options = value.split(' ')
                case 'unbound':
                    # (pid 520) is running...
                    # is stopped
                    re_pid = r'\(pid (\d+)\).*'
                    re_status = r'.*(is \w+)'
                    if pid := re_match(re_pid, value):
                        pid = pid[1]
                    if status := re_match(re_status, value):
                        status = status[1]
                case _:
                    pass

        try:
            return cls(
                modules=modules,
                options=options,
                pid=pid,
                status=status,
                threads=threads,
                uptime=uptime,
                verbosity=verbosity,
                version=version,
            )
        except NameError as e:
            if debug.enabled:
                print(f'name error: {e}')
            return


def parse_unbound_status(string_table: StringTable) -> UnboundStatus | None:
    return UnboundStatus.parse(string_table)


register.agent_section(
    name="unbound_status",
    parse_function=parse_unbound_status,
)


def discover_unbound_status(section: UnboundStatus) -> DiscoveryResult:
    print(section)
    yield Service()


# UnboundStatus(
#     modules=['subnet', 'validator', 'iterator'],
#     options=['reuseport', 'control(ssl)'],
#     pid=520
#     status='(pid 520) is running',
#     threads=1,
#     uptime=1759,
#     verbosity=0,
#     version='1.13.1',
# )
def check_unbound_status(params: Mapping[str, Any], section: UnboundStatus, ) -> CheckResult:
    if section.status != 'is running':
        yield Result(state=State(params['status_not_running']), summary=f'Status: {section.status}')
        return

    yield Result(state=State.OK, summary=f'Status: {section.status}')
    yield from check_levels(
        value=section.uptime,
        label='Uptime',
        metric_name='uptime',
        render_func=render.timespan
    )
    yield Result(state=State.OK, summary=f'Version: {section.version}')


register.check_plugin(
    name="unbound_status",
    service_name="Unbound",
    discovery_function=discover_unbound_status,
    check_function=check_unbound_status,
    check_default_parameters={
        'status_not_running': 2
    },
    check_ruleset_name="unbound_status",
)
