#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-07-24
# File  : unbound_type_stats.py

# based on the unbound plugin by Jan-Philipp Litza (PLUTEX) <jpl@plutex.de>.
# and the work of ruettimann[at]init7[dot]net

# https://nlnetlabs.nl/projects/unbound/about/

from collections.abc import Mapping, Sequence, MutableSequence
from time import time as now_time
from typing import (
    List,
    Tuple,
)

from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    CheckResult,
    DiscoveryResult,
)

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    GetRateError,
    Result,
    Service,
    State,
    check_levels,
    get_rate,
    get_value_store,
    register,
)

from cmk.base.plugins.agent_based.utils.unbound import (
    UnboundSection,
    render_qps,
)

# https://www.iana.org/assignments/dns-parameters/dns-parameters.xhtml
ANSWER_TYPES_MOST_COMMON: List[str] = [
    'A',
    'AAAA',
    'ANY',
    'CNAME',
    'DNSKEY',
    'DS',
    'HINFO',
    'HTTPS',
    'MX',
    'NAPTR',
    'NS',
    'PTR',
    'RRSIG',
    'SOA',
    'SRV',
    'SVCB',
    'TXT',
    # 'TYPE65',  # in summary for HTTPS
]
ANSWER_TYPES_COMMON: List[str] = [
    'AFSDB',
    'APL',
    'CAA',
    'CDNSKEY',
    'CDS',
    'CERT',
    'CSYNC',
    'DHCID',
    'DLV',
    'DNAME',
    'EUI48',
    'EUI64',
    'HIP',
    'IPSECKEY',
    'KEY',
    'KX',
    'LOC',
    'NSEC',
    'NSEC3',
    'NSEC3PARAM',
    'OPENPGPKEY',
    'RP',
    'SIG',
    'SMIMEA',
    'SSHFP',
    'TA',
    'TKEY',
    'TLSA',
    'TSIG',
    'URI',
    'WALLET',
    'ZONEMD',
]
ANSWER_TYPES_OBSOLETE: List[str] = [
    'A6',
    'ATMA',
    'DOA',
    'EID',
    'GID',
    'GPOS',
    'ISDN',
    'L32',
    'L64',
    'LP',
    'MAILA',
    'MAILB',
    'MB',
    'MD',
    'MF',
    'MG',
    'MINFO',
    'MR',
    'NB',
    'NBSTAT',
    'NID',
    'NIMLOC',
    'NINFO',
    'NSAP',
    'NSAP-PTR',
    'NULL',
    'NXT',
    'PX',
    'RKEY',
    'RT',
    'SINK',
    'SPF',
    'TALINK',
    'UID',
    'UINFO',
    'UNSPEC',
    'WKS',
    'X25',
]

ANSWER_TYPES: List[str] = ANSWER_TYPES_MOST_COMMON + ANSWER_TYPES_COMMON + ANSWER_TYPES_OBSOLETE

SUM_TYPES: Mapping[str: Sequence[str]] = {
    'AAAA': ['AAAA', 'A6'],
    'HTTPS': ['HTTPS', 'TYPE65'],
}


def discover_unbound_type_stats(section: UnboundSection) -> DiscoveryResult:
    if 'num.query.type.A' in section:
        yield Service()


def yield_answer(
        answer: str,
        levels_upper: Tuple[int, int] | None,
        total: float,
        value: float,
        notice: bool = True,
        metric: bool = True,
):
    now = now_time()
    try:
        rate = get_rate(
            get_value_store(),
            f'unbound_type_stats_{answer}',
            now,
            value,
            raise_overflow=True,
        )
    except GetRateError as e:
        yield Result(state=State.OK, notice=str(e))
        return
    else:
        # if rate == 0:
        #     return
        if levels_upper is not None and len(levels_upper) == 3:
            # levels on the ratio of answers
            levels_upper = (
                levels_upper[0] * total,
                levels_upper[1] * total,
            )
        # if levels_upper:
        #     notice = False
        metric_name = f'unbound_type_stats_{answer.replace("-", "_").replace(" ", "_")}' if metric else None
        yield from check_levels(
            value=rate,
            levels_upper=levels_upper,
            metric_name=metric_name,
            render_func=render_qps,
            label=answer,
            notice_only=notice,
        )


# ToDo: values are converted to answers/s, total is an absolute value -> should be changed to rate/1?
def check_unbound_type_stats(params: Mapping, section: UnboundSection) -> CheckResult:
    key_prefix = 'num.query.type.'
    total_common: float = 0.0
    total_obsolete: float = 0.0
    other_value: float = 0.0
    other_types_found: MutableSequence[str] = []

    for sum_type in SUM_TYPES:
        summary = 0.0
        for answer_type in SUM_TYPES[sum_type]:
            key = f'{key_prefix}{answer_type}'
            summary += section.get(key, 0.0)
            try:
                section.pop(key)
            except KeyError:
                pass

        if summary > 0.0:
            section.update({f'{key_prefix}{sum_type}': summary})

    section = dict(sorted(section.items()))

    total = sum(
        value for key, value in section.items()
        if key.startswith(key_prefix)
    )

    for key, value in section.items():
        if not key.startswith(key_prefix):
            continue
        answer = key[len(key_prefix):]

        if answer in ANSWER_TYPES:
            if answer in ANSWER_TYPES_COMMON:
                total_common += value
            elif answer in ANSWER_TYPES_OBSOLETE:  #
                total_obsolete += value

            yield from yield_answer(
                answer=answer,
                total=total,
                value=value,
                levels_upper=params.get(f'levels_upper_{answer}')
            )
        else:
            other_types_found.append(answer)
            other_value += value
            yield from yield_answer(
                answer=answer,
                total=total,
                value=value,
                levels_upper=None,
                metric=False,
            )

    if other_types_found:
        answer = 'others'
        yield from yield_answer(
            answer=answer,
            total=total,
            value=other_value,
            levels_upper=params.get(f'levels_upper_{answer}')
        )
        answer_types = ', '.join(other_types_found)
        yield Result(
            state=State(params.get('state_other_type', 0)),
            notice=f'Other DNS answer typ(e) found: {answer_types}'
        )

    for answer, value in [
        ('Total', total),
        ('Total_common', total_common),
        ('Total_obsolete', total_obsolete),
    ]:
        if value > 0:
            yield from yield_answer(
                answer=answer,
                total=total,
                value=value,
                levels_upper=params.get(f'levels_upper_{answer}'),
                notice=False,
            )


register.check_plugin(
    name='unbound_type_stats',
    service_name='Unbound Type Stats',
    sections=['unbound'],
    discovery_function=discover_unbound_type_stats,
    check_function=check_unbound_type_stats,
    check_default_parameters={
    },
    check_ruleset_name='unbound_type_stats',
)
